module vga11 (clk,r,g,b,hs,vs,vgaclk,       // vga interface
             addr,datain,dataout,wr,rd,ready,cs,      // cpu interface
             sram_addr,sram_ce_n,sram_dq,sram_lb_n,sram_oe_n,sram_ub_n,sram_we_n,
				 debug,hbl,vbl,dli,
				 left,right,audiostb,empty,
				 leds,hex0,hex1,
				 sw,key,ledin,audioin
				 
				 
				 );// sram interface); 
				
// system driver for DE2-115 retromachine
// ver 0.98 beta --- 2014.12.06
// (c) Piotr Kardasz pik@we.pb.edu.pl; pik33@o2.pl
// This is free software, license: LGPL v.2.0 or newer
// ***** Chaotic code and comments warning ******
// ***** This will be corrected until 1.0 **************** 

output reg [23:0] left;
output reg [23:0] right;
input wire audiostb;
output reg empty;
output reg [31:0] leds,hex0,hex1;
input wire [17:0] sw;
input wire [3:0] key;
input wire [7:0] ledin;
output reg [7:0] r,g,b;
output reg hs,vs;
output wire vgaclk; 
input wire clk;  //input clock=2x vgaclk
output reg ready;
input wire cs;
output reg[7:0] debug;

reg [9:0] e,ee,eee,eeee;
reg f,oldwr,oldwr2;
reg [7:0] brr,brb,brg;
reg [19:0] caddr,caddr2;
reg [31:0] lcnt;
reg [0:2] lstate,pstate; // 00 vis, 01 fp, 10 sync, 11 bp 100 border
reg [11:0] pcnt;
reg [31:0] fcnt; //frame counter
reg[5:0] sc3;
input wire [20:0] addr;
input wire [31:0] datain;
output reg [31:0] dataout;
input wire wr,rd;
output reg [19:0] sram_addr;
inout wire [15:0] sram_dq;
output wire sram_ce_n;
output reg sram_lb_n;
output wire sram_oe_n;
output reg sram_ub_n;
output reg sram_we_n;
output reg hbl,vbl,dli;
input wire [31:0] audioin;
reg[7:0] rh,bh,gh;
reg spr,spr2;
reg [15:0] sramdqout,dataout2,dataout3;
reg[31:0] datain2,olddatain;
reg[63:0] datain3;
reg [20:0] addr2,sram_addr2,addr3;
reg[3:0] pcnt_mod,pcnt_mod2;
reg[7:0] s1p,s1c,s2c,s3c,s4c,s5c,s6c,s1c1,s2c1,s3c1,s4c1,s5c1,s6c1,sll;
reg s1c2,s2c2,s3c2,s4c2,s5c2,s6c2;
reg[15:0] pixels3, pixels2;
reg[31:0] pixels,bb,cc,ccc,ss,linebase,dl;
reg[1:0] hmode,vmode,sm1x,sm1y,sp1,sm2x,sm2y,sp2,sm3x,sm3y,sp3,sm4x,sm4y,sp4,rhl,ghl,bhl;
reg[1:0] sm5x,sm5y,sp5,sm6x,sm6y,sp6;
reg wr2,wr3,rd2,rd3,oldrd,dla;
reg[2:0]wstate;
reg[31:0] s1x,s1y,s2x,s2y,s3x,s3y,s4x,s4y,s5x,s5y,s6x,s6y;
reg[31:0] s6l1,s5l1,s3l1,s4l1,s1l1,s2l1,saddr,sdata;
reg[31:0] s1l[8],s2l[8],s3l[8],s4l[8],s5l[8],s6l[8];
reg[7:0] palmem;
reg[11:0] pcnt_char;
reg[2:0] paulanum;
reg[10:0]vga1;
reg[31:0]r42;
//reg add;
(* ramstyle = "M9K" *) reg [31:0] pallette[1024]; // pallette registers
(* ramstyle = "M9K" *) reg [7:0] fontdef[2048];  // character generator
(* ramstyle = "M9K" *) reg [31:0] sprite[1536]; // sprite generator

reg[31:0] regs[64];  // internal regs
reg[31:0] vga[13];  // internal regs
reg signed [31:0] paula[8];
reg signed [31:0] paula2[8];


// 0 - frame counter
// 1 - display start address
// 2 - mode
// 3 - border color
// 4 - pallette bank
// 5 - pallette selector
// 6 - display list start address; bit 31: dl active
// 7 - horizontal scroll
// 8 - line length
// 9..14 - sprites
// 15 - collision
// 16..19 - paula output regs
// 20..23 - paula freq regs
// 24..27 - paula start regs
// 28..31 - paula len regs
// 32..35 - paula vol regs



// 8-channel Paula-type chip
// 0..7 - paula counter regs; bits 31..12-> ram addressing; freq regs are in regs space[24..31] because they has to be written from system
// 8..15 - paula currend sample addr
// 16..23 - paula current sample


// display list 
// [31] - entry valid
// [30] - generate dli
// [29..22] palette index
// [21..18] mode
// [17..0] line start address; 

reg [31:0] buff [1024];
reg [31:0] adata;
reg [9:0] aaddr;
reg oldaudiostb;

reg signed[15:0] paulal,paular;
reg signed[15:0] dacl,dacr;


initial
  begin
//  $readmemh("sinus.hex",buff);
  aaddr<=0;
  empty<=0;
  end
  
  
// -----------------------paula

always@ (posedge audiostb)

begin
 
  r42<=audioin;
// len regs bit 31=1 - one shot 
  paula[0]=paula[0]+regs[20]; // inc counter
  if ((paula[0][31:12]>=regs[28][19:0]) && (regs[28][31]==0)) paula[0][31:12]=paula[0][31:12]-regs[28][19:0]; // if>len, corrrect
  else if ((paula[0][31:12]>=regs[28][19:0]) && (regs[28][31]==1)) paula[0][31:12]=0; // if>len, corrrect
  paula[4]<=regs[24]+paula[0][31:12];
  
  paula[1]=paula[1]+regs[21]; // inc counter
  if ((paula[1][31:12]>=regs[29][19:0]) && (regs[29][31]==0)) paula[1][31:12]=paula[1][31:12]-regs[29][19:0]; // if>len, corrrect
  else if ((paula[1][31:12]>=regs[29][19:0]) && (regs[29][31]==1)) paula[1][31:12]=0; 
  paula[5]<=regs[25]+paula[1][31:12];
  
  paula[2]=paula[2]+regs[22]; // inc counter
  if ((paula[2][31:12]>=regs[30][19:0]) && (regs[30][31]==0)) paula[2][31:12]=paula[2][31:12]-regs[30][19:0]; // if>len, corrrect
  else if ((paula[2][31:12]>=regs[30][19:0]) && (regs[30][31]==1)) paula[3][31:12]=0; 
  paula[6]<=regs[26]+paula[2][31:12];
  
  paula[3]=paula[3]+regs[23]; // inc counter
  if ((paula[3][31:12]>=regs[31][19:0]) && (regs[31][31]==0)) paula[3][31:12]=paula[3][31:12]-regs[31][19:0]; // if>len, corrrect
  else if ((paula[3][31:12]>=regs[31][19:0]) && (regs[31][31]==1)) paula[3][31:12]=0; 
  paula[7]<=regs[27]+paula[3][31:12];

end  
  
reg signed[15:0] lvol,rvol,lsample,rsample;
reg signed[31:0] lres,rres;
reg [1:0] chn;
  
  
 
  
always@(posedge clk)
   begin

  oldaudiostb<=audiostb; 

  paulal<=regs[16][31:16]+regs[17][31:16]+regs[18][31:16]+regs[19][31:16]+regs[42][31:16];
  paular<=regs[16][15:0]+regs[17][15:0]+regs[18][15:0]+regs[19][15:0]+regs[42][15:0];

  if ((audiostb==1) && (oldaudiostb==0)) // posedge audiostb
     begin
     adata<=buff[aaddr];
	  dacl=adata[31:16];
	  dacr=adata[15:0];
     aaddr<=aaddr+1;
	  empty<=aaddr[9];
//     if (aaddr==128) empty[0]<=1;
//     if (aaddr==0) empty[1]<=1;
	  left={dacl+paulal,8'h00};
	  right={dacr+paular,8'h00};
	  oldaudiostb<=audiostb;
	  end
  end

initial
  begin
  $readmemh("st4font.hex",fontdef);
  regs[45]<=0;   
  regs[47]<=0;  
  regs[43]<=128;
  regs[8]<=224;
  vga[0]<=96;
  vga[1]<=159;
  vga[2]<=1953;
  vga[3]<=2016;
  vga[4]<=2064;
  vga[5]<=2080;
  vga[6]<=25;
  vga[7]<=65;
  vga[8]<=1185;
  vga[9]<=1225;
  vga[10]<=1228;
  vga[11]<=1234;
  vga[12]<=2;
  end


  
assign vgaclk=clk;
assign sram_ce_n=0;
assign sram_oe_n=0;
 

// ------------------------------------------- 
// bidir assignment
// ------------------------------------------- 

assign sram_dq = (sram_we_n==0) ? sramdqout : 16'bz;

//---------------------------------------------  
// debug  
//---------------------------------------------  
  
always@(posedge clk) 
 
begin 
//debug[3:0]<=paulanum;
//debug[7:5]<=chn;
end



//---------------------------------------------
// main loop
//--------------------------------------------- 
 
always@(posedge clk)

  begin
 
 
 	leds<=regs[36];
   hex0<=regs[37];
   hex1<=regs[38];
   regs[39][17:0]<=sw;
	regs[39][31:28]<=key;
   regs[40]<=ledin;
   if (regs[41][0]==0) regs[36][31:24]<=ledin;	

 
 
  chn<=chn+1;
    begin
    lvol=regs[32+chn][31:16];
    rvol=regs[32+chn][15:0];
    lsample=paula2[chn][31:16];
    rsample=paula2[chn][15:0];
    lres=lsample*lvol;
    rres=rsample*rvol;
    if (regs[32+chn]!=0)
      begin
	   regs[16+chn][31:16]<=lres[31:16];
	   regs[16+chn][15:0]<=rres[31:16];
	  end
  end
 

 
// need to read the sprite memory in the simplest way to make Quartus infer it as ram blocks
  sdata<=sprite[saddr];
  
  
  oldwr<=wr;
  if ((oldwr==0) && (wr==1) && cs) wr2<=1; else wr2<=0;
  wr3<=wr2;
  oldrd<=rd;
  if ((oldrd==0) && (rd==1) && cs) rd2<=1; else rd2<=0;
  rd3<=rd2;

  pcnt<=pcnt+1;
  pcnt_mod<=(regs[7]+pcnt+1)&15; 
  pcnt_char<=(regs[7]+pcnt-/*143*/vga[1]+16)>>4;  
  pcnt_mod2<=pcnt_mod-1;  
  if ((pcnt<128) || (pcnt>1952)) pcnt_char[11]<=1; else pcnt_char[11]<=0;
 
 
  case (pcnt)
    17:dli<=0;
    19:if ((regs[6][31]==1)&&(dl[31]==1)) regs[4]<={24'b0,dl[29:22]}; else regs[4]<=regs[4];
	 20:if ((regs[6][31]==1)&&(dl[31]==1)) regs[2]<={28'b0,dl[21:18]}; else regs[2]<=regs[2];
	 21:dl<=0;
    34:bb<=regs[3];
    35:vmode<=regs[2][1:0];
	 36:hmode<=regs[2][3:2];
	 2057:if ((regs[6][31]==1)&&(dl[31]==1) && (dl[17:0] != 18'b11_1111_1111_1111_1111)) dla<=1; else dla<=0;
	 2056:if ((regs[6][31]==1)&&(dl[31]==1)) dli<=1;
  endcase   
  
// sprite machine registers read
  
  s1x<=regs[9][13:0];
  s1y<=regs[9][27:16];
  sm1x<=regs[9][15:14];
  sm1y<=regs[9][31:30];
  sp1<=regs[9][29:28];
  
  s2x<=regs[10][13:0];
  s2y<=regs[10][27:16];
  sm2x<=regs[10][15:14];
  sm2y<=regs[10][31:30];
  sp2<=regs[10][29:28];  
  
  s3x<=regs[11][13:0];
  s3y<=regs[11][27:16];
  sm3x<=regs[11][15:14];
  sm3y<=regs[11][31:30];
  sp3<=regs[11][29:28];
  
  s4x<=regs[12][13:0];
  s4y<=regs[12][27:16];
  sm4x<=regs[12][15:14];
  sm4y<=regs[12][31:30];
  sp4<=regs[12][29:28];  
  
  s5x<=regs[13][13:0];
  s5y<=regs[13][27:16];
  sm5x<=regs[13][15:14];
  sm5y<=regs[13][31:30];
  sp5<=regs[13][29:28];
  
  s6x<=regs[14][13:0];
  s6y<=regs[14][27:16];
  sm6x<=regs[14][15:14];
  sm6y<=regs[14][31:30];
  sp6<=regs[14][29:28];  
  

// sprite data read
  
  if (pcnt>=1951) saddr<=(pcnt-1951)+(((lcnt+1-s1y)>>sm1y)<<3);
  if ((pcnt>=1953) && (pcnt<=1960)) s1l[pcnt-1953]<=sdata;	 
  if (pcnt>=1961) saddr<=(pcnt-1961)+(((lcnt+1-s2y)>>sm2y)<<3)+256;
  if ((pcnt>=1963) && (pcnt<=1970)) s2l[pcnt-1963]<=sdata;	 
  if (pcnt>=1971) saddr<=(pcnt-1971)+(((lcnt+1-s3y)>>sm3y)<<3)+512;
  if ((pcnt>=1973) && (pcnt<=1980)) s3l[pcnt-1973]<=sdata;	 
  if (pcnt>=1981) saddr<=(pcnt-1981)+(((lcnt+1-s4y)>>sm4y)<<3)+768;
  if ((pcnt>=1983) && (pcnt<=1990)) s4l[pcnt-1983]<=sdata;	 
  if (pcnt>=1991) saddr<=(pcnt-1991)+(((lcnt+1-s5y)>>sm5y)<<3)+1024;
  if ((pcnt>=1993) && (pcnt<=2000)) s5l[pcnt-1993]<=sdata;	 
  if (pcnt>=2001) saddr<=(pcnt-2001)+(((lcnt+1-s6y)>>sm6y)<<3)+1280;
  if ((pcnt>=2003) && (pcnt<=2010)) s6l[pcnt-2003]<=sdata;	 

// paula data read



  
// ------------------------------------------- 
// sram rw state machine
// -------------------------------------------
  
// new state calculation
  case (wstate) 
  
    0: begin
	   addr2<=addr;
		if (((wr2==1) || (wr3==1)) && (rd2==0) && (rd3==0) && (addr[19]==0)) 
			 begin 
			 addr3=addr+regs[47]; wstate<=1; 
			 if (regs[45][31]==0) datain3={regs[46],datain}>>regs[45][4:0];
		    else datain3=({datain,regs[46]}<<regs[45][4:0])>>32; 
			 end

		else if (((wr2==1) || (wr3==1)) && (rd2==0) && (rd3==0) && (addr[19]==1)) wstate<=3;
		else if (((rd2==1) || (rd3==1)) && (wr2==0) && (wr3==0) && (addr[19]==0)) begin addr3=addr+regs[47]; wstate<=4; end
		else if (((rd2==1) || (rd3==1)) && (wr2==0) && (wr3==0) && (addr[19]==1)) wstate<=5;
		else wstate<=0;
		datain2<=datain;
		end
    1: begin  
		if (pcnt_mod2==0) wstate<=6; else wstate<=1; 
		sramdqout<=datain3[31:16];
		sram_addr2<={addr3,1'b0};
      end 
    6: begin	 	
	   regs[46]<=datain2;
	   case (pcnt_mod)
		1,2,3:
		    begin
	       sram_addr<=sram_addr2;
	       sram_we_n<=0;    
          sram_lb_n<=0;
          sram_ub_n<=0;  
 	       end		
		4: 
		    begin
			 sram_addr<=sram_addr2;
  	       sram_we_n<=0;    
          sram_lb_n<=1;
          sram_ub_n<=1;  
	       end 	
		5,6,7:   	   
		    begin
		    sram_addr<=sram_addr2+1;
  	       sram_we_n<=0;    
          sram_lb_n<=0;
          sram_ub_n<=0;  
   	    sramdqout<=datain3[15:0];
			 end
   	8:
		    begin		    
			 sram_addr<=sram_addr2+1;
			 sram_we_n<=1;    
          sram_lb_n<=1;
          sram_ub_n<=1;  
			 wstate<=0;
          end
		default:
			wstate<=0;
		endcase	
     end
  
	4:	begin
    	if (pcnt_mod2==0) wstate<=2; else wstate<=4;
		ready<=0;
		sram_addr2<={addr3,1'b0};
      end		
	2:begin	
	   case (pcnt_mod2)
      1,2,3:
		    begin
	       sram_addr<=sram_addr2;
	       sram_we_n<=1;    
          sram_lb_n<=0;
          sram_ub_n<=0;  
 	       end
		4:
         begin
   	   sram_addr<=sram_addr2+1;
	      dataout3<=sram_dq;
  	      sram_we_n<=1;    
         sram_lb_n<=0;
         sram_ub_n<=0;  
	      end 	 

	   5,6,7:begin
   	   sram_addr<=sram_addr2+1;
  	      sram_we_n<=1;    
         sram_lb_n<=0;
         sram_ub_n<=0;  
	      end		
		
   	8:
         begin
	      dataout2<=sram_dq;
	      end 	 			
		9:
		   begin
		   dataout<={dataout3,dataout2};	 
			ready<=1;
	      wstate<=0;
			end
	   default:
         wstate<=0;
	      		
			
		endcase	
      end
		
   3: begin  
      wstate<=0;
      if (addr[15:12]==4'b0000) pallette[addr2[9:0]]<=datain; 
		else if (addr[15:12]==4'b0001)fontdef[addr2[10:0]]<=datain[7:0];
		else if (addr[15:12]==4'b0010)sprite[addr2[10:0]]<=datain;
		else if (addr[15:12]==4'b0011)regs[addr2[10:0]]<=datain;
		else if (addr[15:12]==4'b1111) vga[addr2[3:0]]<=datain;
		else buff [addr2[9:0]]<=datain;   	
		end
		
	5: begin	

  	   wstate<=0;
	   if (addr[15:12]==4'b0000) dataout<=pallette[addr2[9:0]]; 
		else if (addr[15:12]==4'b0001) dataout<={24'b0,fontdef[addr2[10:0]]};
		else if (addr[15:12]==4'b0010) dataout<=sprite[addr2[10:0]];
		else if (addr[15:12]==4'b0011) dataout<=regs[addr2[10:0]];
      else if (addr[15:12]==4'b1101) dataout<=paula[addr2[2:0]];
      else if (addr[15:12]==4'b1110) dataout<=r42;
		else if (addr[15:12]==4'b1111) dataout<=vga[addr2[3:0]];
		else dataout<=buff [addr[9:0]];
	   end
	 
   default: wstate<=0;
	
	endcase
	
//----- end of sram interface state machine

//----- vga 
  if (pcnt_mod==8)	 
    begin

    end
	 
  if (pcnt_mod==9)	  
    begin
	 sram_addr<=caddr;
	 sram_we_n<=1;    
    sram_lb_n<=0;
    sram_ub_n<=0;  
	 end 
	 
  if (pcnt_mod==12)	 	 
    begin
	 sram_addr<=caddr+1;
	 pixels2<=sram_dq;
	 sram_we_n<=1;    
    sram_lb_n<=0;
    sram_ub_n<=0;  
	 end 

  if (pcnt_mod==15)	 	 
    begin
	 pixels3=sram_dq;
 	 pixels<={pixels2,pixels3};
	 end 	 
	 
  if (pcnt_mod==0) 
    begin

//	 pixels={pixels2,pixels3};
	 end
	 
	 
  if (pcnt_mod==1) 
    begin  
 	 if ((paulanum!=4) && (pcnt<108)) paulanum<=paulanum+1;
	 if ((paulanum<4)) paula2[paulanum]<=pixels;
	 if (pcnt>2032) dl<=pixels;

    end


  if (regs[5][31]==1)
    begin
    if (pcnt==regs[5][11:0])
	   begin
	   palmem<=regs[4][7:0];
		regs[4]<=regs[4]+regs[5][27:20];
		end
    if (pcnt==1952) regs[4][7:0]<=palmem;
    end
	 
//---------------------sprite 1
  
  if (pcnt==s1x-1) s1l1<=s1l[0];
  if (pcnt==s1x+(4<<sm1x)-1) s1l1<=s1l[1];
  if (pcnt==s1x+(8<<sm1x)-1) s1l1<=s1l[2];
  if (pcnt==s1x+(12<<sm1x)-1) s1l1<=s1l[3];  
  if (pcnt==s1x+(16<<sm1x)-1) s1l1<=s1l[4];
  if (pcnt==s1x+(20<<sm1x)-1) s1l1<=s1l[5];
  if (pcnt==s1x+(24<<sm1x)-1) s1l1<=s1l[6];
  if (pcnt==s1x+(28<<sm1x)-1) s1l1<=s1l[7];
  
  if ((pcnt>=s1x) && (pcnt<s1x+(32<<sm1x)) && (lcnt>=s1y) && (lcnt<s1y+(32<<sm1y))) 
    begin
	 if ((((pcnt-s1x)>>sm1x)&3)==0) begin  s1c=s1l1[31:24]; s1c1<=s1l1[31:24]; end
	 if ((((pcnt-s1x)>>sm1x)&3)==1) begin  s1c=s1l1[23:16]; s1c1<=s1l1[23:16]; end
	 if ((((pcnt-s1x)>>sm1x)&3)==2) begin  s1c=s1l1[15:8]; s1c1<=s1l1[15:8]; end
	 if ((((pcnt-s1x)>>sm1x)&3)==3) begin  s1c=s1l1[7:0]; s1c1<=s1l1[7:0]; end
	 end
	else begin s1c1<=0; s1c=0; end
   s1c2<=|s1c;

	
//------------------sprite 2

  if (pcnt==s2x-1) s2l1<=s2l[0];
  if (pcnt==s2x+(4<<sm2x)-1) s2l1<=s2l[1];
  if (pcnt==s2x+(8<<sm2x)-1) s2l1<=s2l[2];
  if (pcnt==s2x+(12<<sm2x)-1) s2l1<=s2l[3];  
  if (pcnt==s2x+(16<<sm2x)-1) s2l1<=s2l[4];
  if (pcnt==s2x+(20<<sm2x)-1) s2l1<=s2l[5];
  if (pcnt==s2x+(24<<sm2x)-1) s2l1<=s2l[6];
  if (pcnt==s2x+(28<<sm2x)-1) s2l1<=s2l[7];
  
  if ((pcnt>=s2x) && (pcnt<s2x+(32<<sm2x)) && (lcnt>=s2y) && (lcnt<s2y+(32<<sm2y))) 
    begin
	 if ((((pcnt-s2x)>>sm2x)&3)==0) begin  s2c=s2l1[31:24]; s2c1<=s2l1[31:24]; end
	 if ((((pcnt-s2x)>>sm2x)&3)==1) begin  s2c=s2l1[23:16]; s2c1<=s2l1[23:16]; end
	 if ((((pcnt-s2x)>>sm2x)&3)==2) begin  s2c=s2l1[15:8]; s2c1<=s2l1[15:8]; end
	 if ((((pcnt-s2x)>>sm2x)&3)==3) begin  s2c=s2l1[7:0]; s2c1<=s2l1[7:0]; end
	 end
	else begin s2c1<=0; s2c=0; end
	s2c2<=|s2c;

//------------------sprite 3

  if (pcnt==s3x-1) s3l1<=s3l[0];
  if (pcnt==s3x+(4<<sm3x)-1) s3l1<=s3l[1];
  if (pcnt==s3x+(8<<sm3x)-1) s3l1<=s3l[2];
  if (pcnt==s3x+(12<<sm3x)-1) s3l1<=s3l[3];  
  if (pcnt==s3x+(16<<sm3x)-1) s3l1<=s3l[4];
  if (pcnt==s3x+(20<<sm3x)-1) s3l1<=s3l[5];
  if (pcnt==s3x+(24<<sm3x)-1) s3l1<=s3l[6];
  if (pcnt==s3x+(28<<sm3x)-1) s3l1<=s3l[7];
  
  if ((pcnt>=s3x) && (pcnt<s3x+(32<<sm3x)) && (lcnt>=s3y) && (lcnt<s3y+(32<<sm3y))) 
    begin
	 if ((((pcnt-s3x)>>sm3x)&3)==0) begin  s3c=s3l1[31:24]; s3c1<=s3l1[31:24]; end
	 if ((((pcnt-s3x)>>sm3x)&3)==1) begin  s3c=s3l1[23:16]; s3c1<=s3l1[23:16]; end
	 if ((((pcnt-s3x)>>sm3x)&3)==2) begin  s3c=s3l1[15:8]; s3c1<=s3l1[15:8]; end
	 if ((((pcnt-s3x)>>sm3x)&3)==3) begin  s3c=s3l1[7:0]; s3c1<=s3l1[7:0]; end
	 end
	else begin s3c1<=0; s3c=0; end
	s3c2<=|s3c;
	
//------------------sprite 4

  if (pcnt==s4x-1) s4l1<=s4l[0];
  if (pcnt==s4x+(4<<sm4x)-1) s4l1<=s4l[1];
  if (pcnt==s4x+(8<<sm4x)-1) s4l1<=s4l[2];
  if (pcnt==s4x+(12<<sm4x)-1) s4l1<=s4l[3];  
  if (pcnt==s4x+(16<<sm4x)-1) s4l1<=s4l[4];
  if (pcnt==s4x+(20<<sm4x)-1) s4l1<=s4l[5];
  if (pcnt==s4x+(24<<sm4x)-1) s4l1<=s4l[6];
  if (pcnt==s4x+(28<<sm4x)-1) s4l1<=s4l[7];
  
  if ((pcnt>=s4x) && (pcnt<s4x+(32<<sm4x)) && (lcnt>=s4y) && (lcnt<s4y+(32<<sm4y))) 
    begin
	 if ((((pcnt-s4x)>>sm4x)&3)==0) begin  s4c=s4l1[31:24]; s4c1<=s4l1[31:24]; end
	 if ((((pcnt-s4x)>>sm4x)&3)==1) begin  s4c=s4l1[23:16]; s4c1<=s4l1[23:16]; end
	 if ((((pcnt-s4x)>>sm4x)&3)==2) begin  s4c=s4l1[15:8]; s4c1<=s4l1[15:8]; end
	 if ((((pcnt-s4x)>>sm4x)&3)==3) begin  s4c=s4l1[7:0]; s4c1<=s4l1[7:0]; end
	 end
	else begin s4c1<=0; s4c=0; end
   s4c2<=|s4c;	

    
//---------------------sprite 5
  
  if (pcnt==s5x-1) s5l1<=s5l[0];
  if (pcnt==s5x+(4<<sm5x)-1) s5l1<=s5l[1];
  if (pcnt==s5x+(8<<sm5x)-1) s5l1<=s5l[2];
  if (pcnt==s5x+(12<<sm5x)-1) s5l1<=s5l[3];  
  if (pcnt==s5x+(16<<sm5x)-1) s5l1<=s5l[4];
  if (pcnt==s5x+(20<<sm5x)-1) s5l1<=s5l[5];
  if (pcnt==s5x+(24<<sm5x)-1) s5l1<=s5l[6];
  if (pcnt==s5x+(28<<sm5x)-1) s5l1<=s5l[7];
  
  if ((pcnt>=s5x) && (pcnt<s5x+(32<<sm5x)) && (lcnt>=s5y) && (lcnt<s5y+(32<<sm5y))) 
    begin
	 if ((((pcnt-s5x)>>sm5x)&3)==0) begin  s5c=s5l1[31:24]; s5c1<=s5l1[31:24]; end
	 if ((((pcnt-s5x)>>sm5x)&3)==1) begin  s5c=s5l1[23:16]; s5c1<=s5l1[23:16]; end
	 if ((((pcnt-s5x)>>sm5x)&3)==2) begin  s5c=s5l1[15:8]; s5c1<=s5l1[15:8]; end
	 if ((((pcnt-s5x)>>sm5x)&3)==3) begin  s5c=s5l1[7:0]; s5c1<=s5l1[7:0]; end
	 end
	else begin s5c1<=0; s5c=0; end
	s5c2<=|s5c;

//------------------sprite 6

  if (pcnt==s6x-1) s6l1<=s6l[0];
  if (pcnt==s6x+(4<<sm6x)-1) s6l1<=s6l[1];
  if (pcnt==s6x+(8<<sm6x)-1) s6l1<=s6l[2];
  if (pcnt==s6x+(12<<sm6x)-1) s6l1<=s6l[3];  
  if (pcnt==s6x+(16<<sm6x)-1) s6l1<=s6l[4];
  if (pcnt==s6x+(20<<sm6x)-1) s6l1<=s6l[5];
  if (pcnt==s6x+(24<<sm6x)-1) s6l1<=s6l[6];
  if (pcnt==s6x+(28<<sm6x)-1) s6l1<=s6l[7];
  
  if ((pcnt>=s6x) && (pcnt<s6x+(32<<sm6x)) && (lcnt>=s6y) && (lcnt<s6y+(32<<sm6y))) 
    begin
	 if ((((pcnt-s6x)>>sm6x)&3)==0) begin  s6c=s6l1[31:24]; s6c1<=s6l1[31:24]; end
	 if ((((pcnt-s6x)>>sm6x)&3)==1) begin  s6c=s6l1[23:16]; s6c1<=s6l1[23:16]; end
	 if ((((pcnt-s6x)>>sm6x)&3)==2) begin  s6c=s6l1[15:8]; s6c1<=s6l1[15:8]; end
	 if ((((pcnt-s6x)>>sm6x)&3)==3) begin  s6c=s6l1[7:0]; s6c1<=s6l1[7:0]; end
	 end
	else begin s6c1<=0; s6c=0; end
	s6c2<=|s6c;
	
  if ((s1c) && (s2c)) regs[15][0]<=1;
  if ((s1c) && (s3c)) regs[15][1]<=1;
  if ((s1c) && (s4c)) regs[15][2]<=1;
  if ((s1c) && (s5c)) regs[15][3]<=1;
  if ((s1c) && (s6c)) regs[15][4]<=1;
  if ((s2c) && (s3c)) regs[15][5]<=1;
  if ((s2c) && (s4c)) regs[15][6]<=1;
  if ((s2c) && (s5c)) regs[15][7]<=1;
  if ((s2c) && (s6c)) regs[15][8]<=1;
  if ((s3c) && (s4c)) regs[15][9]<=1;
  if ((s3c) && (s5c)) regs[15][10]<=1;
  if ((s3c) && (s6c)) regs[15][11]<=1;
  if ((s4c) && (s5c)) regs[15][12]<=1;
  if ((s4c) && (s6c)) regs[15][13]<=1;
  if ((s5c) && (s6c)) regs[15][14]<=1;

  
	
//  vga[0]<=96;
//  vga[1]<=159;
//  vga[2]<=1953;
//  vga[3]<=2016;
//  vga[4]<=2064;
//  vga[5]<=2080;
//  vga[6]<=25;
//  vga[7]<=65;
//  vga[8]<=1185;
//  vga[9]<=1225;
//  vga[10]<=1228;
//  vga[11]<=1234;
	
//----- end of sprites
	
    // line and pixel counters & state determination
  if (pcnt==vga[0]) begin pstate<=4; hbl<=0; end  // left border
  if (pcnt==vga[1]) pstate<=0; // display
  if (pcnt==vga[2]) begin pstate<=4; end;// right border
  if (pcnt==vga[3]) begin pstate<=1; hbl<=1; end // front porch
  if (pcnt==vga[4]) begin pstate<=2; hs<=vga[12][0]; end// hsync
  if (pcnt==vga[5])          // back porch 
	 begin 
	 vga1<=vga[1]+2;
	 paulanum<=7;
	 pstate<=3; 
    hs<=~hs;
	 pcnt<=17;
	 lcnt<=lcnt+1; 
    regs[0]<=lcnt+1;
//    regs[42]<=r42;
//    ee<=0;
    if ((dl[31]==1) && (regs[6][31]==1) && (dla==1)) linebase<=dl[17:0]<<2; 
    else if (lcnt==65) linebase<=regs[1]<<1;
	 else if (vmode==0)  linebase<=linebase+regs[8];
	 else if ((vmode==1) && (lcnt[0]==1'b1)) linebase<=linebase+regs[8];
	 else if ((vmode==2) && (lcnt[1:0]==2'b01)) linebase<=linebase+regs[8];
	 else if ((vmode==3) && (lcnt[3:0]==4'b0001)) linebase<=linebase+regs[8];

	 else linebase<=linebase;
		
	 if (lcnt==vga[6]) begin vbl<=0; lstate<=4; end   // top border
	 else if (lcnt==vga[7]) begin  lstate<=0; linebase<=(regs[1]<<1); end   // display
	 else if (lcnt==vga[8]) lstate<=4;  // bottom border
	 else if (lcnt==vga[9]) begin lstate<=1; vbl<=1;  end  // front porch v
	 else if (lcnt==vga[10]) begin lstate<=2; vs<=vga[12][1]; end // vsync 
	 else if (lcnt==vga[11]) begin lstate<=3; vs<=~vs; lcnt<=0; fcnt<=fcnt+1;  end // back porch v
	 else begin lstate<=lstate; vs<=vs; end
	 
	 end
 
//    caddr=((lcnt-66)>>mode)*224+(pcnt_char<<1)+(regs[1]<<1);
    if (pcnt_char[11]==0) caddr=linebase+(pcnt_char<<1); 
	 else if (pcnt>=2032) caddr=((regs[6][19:0]+lcnt)<<1);
	 else if ((paulanum<4)) caddr=(paula[4+paulanum])<<1; 

//	 else if (((pcnt>16) && (pcnt<112)))//     || ((pcnt>1952) && (pcnt<2032))) 
//	   begin
//		paulanum=(pcnt-16)>>4;
//		caddr=(paula[8+paulanum])<<1;
//		end
//	 else if ((pcnt>=2032)&&(pcnt_mod==7))caddr<=paula[8];

	 if (pstate==1)
	   begin
		brr=bb[7:0];
		brg=bb[15:8];
		brb=bb[23:16];
      end 
    
    if (((lstate==4)&&((pstate&3)==0)) || ((pstate==4)&&(lstate==0))) // borders
      begin 
      r<=brr; g<=brg; b<=brb;	
      end		 
	 
    else if (lstate==0 && pstate==0) // active display
      begin

      
		if (hmode==0)
        case (pcnt_mod)
          0:e={regs[4][7:0],pixels[31:30]};
	  	    1:e={regs[4][7:0],pixels[29:28]};
		    2:e={regs[4][7:0],pixels[27:26]};
		    3:e={regs[4][7:0],pixels[25:24]};
		    4:e={regs[4][7:0],pixels[23:22]};
		    5:e={regs[4][7:0],pixels[21:20]};
		    6:e={regs[4][7:0],pixels[19:18]};
		    7:e={regs[4][7:0],pixels[17:16]};		            
			 8:e={regs[4][7:0],pixels[15:14]};
	  	    9:e={regs[4][7:0],pixels[13:12]};
		    10:e={regs[4][7:0],pixels[11:10]};
		    11:e={regs[4][7:0],pixels[9:8]};
		    12:e={regs[4][7:0],pixels[7:6]};
		    13:e={regs[4][7:0],pixels[5:4]};
		    14:e={regs[4][7:0],pixels[3:2]};
		    15:e={regs[4][7:0],pixels[1:0]};		  
		  endcase
		else if (hmode==1)  
		   case (pcnt_mod)
  		   0,1:e={regs[4][5:0],pixels[31:28]};
		   2,3:e={regs[4][5:0],pixels[27:24]};
		   4,5:e={regs[4][5:0],pixels[23:20]};
		   6,7:e={regs[4][5:0],pixels[19:16]};  		   
			8,9:e={regs[4][5:0],pixels[15:12]};
		   10,11:e={regs[4][5:0],pixels[11:8]};
		   12,13:e={regs[4][5:0],pixels[7:4]};
		   14,15:e={regs[4][5:0],pixels[3:0]};
       endcase
	
   	else 
		   case (pcnt_mod)
  		   0,1,2,3:e={regs[4][1:0],pixels[31:24]};
		   4,5,6,7:e={regs[4][1:0],pixels[23:16]};
		   8,9,10,11:e={regs[4][1:0],pixels[15:8]};
		   12,13,14,15:e={regs[4][1:0],pixels[7:0]};  		   
        endcase
		  
		  
     eeee<=e;		
		
   if (e>=regs[44]) begin 
		if (s1c!=0)  regs[15][15]<=1; 
		if (s2c!=0)  regs[15][16]<=1; 
		if (s3c!=0)  regs[15][17]<=1; 
		if (s4c!=0)  regs[15][18]<=1; 
	   if (s5c!=0)  regs[15][19]<=1; 
		if (s6c!=0)  regs[15][20]<=1; 
	   end
  
    if (e<regs[43]) 
	   begin
   
//      spr<=(s1c2)|(s2c2)|(s3c2)|(s4c2)|(s5c2)|(s6c2);
		spr<=|{s1c2,s2c2,s3c2,s4c2,s5c2,s6c2};
//		sc3={s6c2,s5c2,s4c2,s3c2,s2c2,s1c2};
      

		if (s1c2)ee<={sp1,s1c1}; 
	   else if (s2c2)  ee<={sp2,s2c1};
		else if (s3c2)  ee<={sp3,s3c1}; 
      else if (s4c2)  ee<={sp4,s4c1};
	   else if (s5c2)  ee<={sp5,s5c1}; 
	   else if (s6c2)  ee<={sp6,s6c1}; 
		else ee<=e; 
  	   end
			
    else ee<=e;
      
  	 spr2<=spr;	
  	 cc<=pallette[ee];
	 ccc<=pallette[eeee]; 
    eee<=eeee;
	 
	 if (eee[7:6]==0) begin rh<=ccc[7:0]; gh<=ccc[15:8]; bh<=ccc[23:16]; end
	 else if (eee[7:6]==1) rh[7:2]<=eee[5:0];
	 else if (eee[7:6]==2) gh[7:2]<=eee[5:0];
	 else bh[7:2]<=eee[5:0];
	 
	 if (pcnt>vga1)
	   begin
	   if (spr2==0)
		  begin  
		  if (hmode!=3) 
		    begin r=ccc[7:0]; g=ccc[15:8]; b=ccc[23:16]; end
		  else 
	       begin 
          if (eee[7:6]==0) begin r=ccc[7:0]; g=ccc[15:8]; b=ccc[23:16]; end
          else begin r=rh; b=bh; g=gh; end 
  	       end
		  end
	   else begin r=cc[7:0]; g=cc[15:8]; b=cc[23:16]; end 
	 end	
	 else begin r<=brr; g<=brg; b<=brb; end		
  end
  else begin r<=0; g<=0; b<=0; end	
  end

endmodule 